/*
# IngPerl - A Ingres Database Interface for Perl 4
#
# Formally sqlperl by Ted Lemon.
# Now developed and maintained by Tim Bunce, timbo@ig.co.uk
# Copyright 1994 Tim Bunce and Ted Lemon
# The latest version is held in ftp.demon.co.uk:/pub/perl/db/ingperl
#
# IngPerl is supplied under the same terms as perl itself,
# e.g., either GNU GPL or Artistic License at your choice.
#
# IngPerl comes with no warranty - it works for me - it may not
# work for you. If it trashes your database I am not responsible!
#
# ingperl.pl should be included in all applications using ingperl
# in order to help ensure that scripts will remain compatible with
# new releases of ingperl.
*/

#include <EXTERN.h>
#include <perl.h>

#include "patchlevel.h"

EXEC SQL INCLUDE SQLCA;
EXEC SQL INCLUDE SQLDA;

int on_error();
int on_warning();
int on_message();
int on_dbevent();

EXEC SQL WHENEVER SQLERROR   CALL on_error;
EXEC SQL WHENEVER SQLWARNING CALL on_warning;
EXEC SQL WHENEVER SQLMESSAGE CALL on_message;
EXEC SQL WHENEVER DBEVENT    CALL on_dbevent;

EXEC SQL BEGIN DECLARE SECTION;
char *SqlTmp;	/* for general use when interfacing with ingres */
int	  SqlInt;	/* ditto */
char *SqlBuf;
EXEC SQL END DECLARE SECTION;

static IISQLDA	*sqlda = (IISQLDA *)0;
static int		connected = 0;
static int		cursor_open = 0;
static int		sql_readonly = 1;
static int		sql_showerrors = 0;
static int		sql_debug = 0;

static int		close_sql_cursor();

/* see on_event function */
enum on_events{
	OE_error,
	OE_warning,
	OE_message,
	OE_dbevent
};

/* ingperl functions (not actually used for many funcs) */
enum userfuncs {
    US_sql,
    US_sql_exec,
    US_sql_fetch,
    US_sql_close,
    US_sql_types,
    US_sql_ingtypes,
	US_sql_nullable,
    US_sql_lengths,
    US_sql_names,
};

/* ingperl variables */
enum uservars {
    UV_sql_error,
    UV_sql_sqlcode,
	UV_sql_version,
	UV_sql_showerrors,
	UV_sql_readonly,
	UV_sql_debug,
	UV_sql_rowcount
};
static int userset();
static int userval();

int
init_sql()
{
	char *file = "ingperl.c";
	int do_sql();
	int do_sql_exec();
	int do_sql_desc();
	int do_sql_fetch();
	int do_sql_close();
    struct ufuncs uf;

    uf.uf_set = userset;
    uf.uf_val = userval;
#define MAGICVAR(name, ix) \
	uf.uf_index=ix, magicname(name, &uf, sizeof uf)

	MAGICVAR("sql_error",      UV_sql_error);
	MAGICVAR("sql_sqlcode",    UV_sql_sqlcode);
	MAGICVAR("sql_version",    UV_sql_version);
	MAGICVAR("sql_showerrors", UV_sql_showerrors);
	MAGICVAR("sql_readonly",   UV_sql_readonly);
	MAGICVAR("sql_debug",      UV_sql_debug);
	MAGICVAR("sql_rowcount",   UV_sql_rowcount);

	make_usub("sql",        US_sql,        do_sql,       file);
	make_usub("sql_exec",   US_sql_exec,   do_sql_exec,  file);
	make_usub("sql_fetch",  US_sql_fetch,  do_sql_fetch, file);
	make_usub("sql_close",  US_sql_close,  do_sql_close, file);

	make_usub("sql_types",    US_sql_types,    do_sql_desc,  file);
	make_usub("sql_ingtypes", US_sql_ingtypes, do_sql_desc,  file);
	make_usub("sql_nullable", US_sql_nullable, do_sql_desc,  file);
	make_usub("sql_lengths",  US_sql_lengths,  do_sql_desc,  file);
	make_usub("sql_names",    US_sql_names,    do_sql_desc,  file);

	return 0;
}


int
do_sql(ix, sp, items)
	int ix;
	register int sp;
	register int items;
{
	STR **st = stack->ary_array + sp;
	register char *tmps;
	register STR *Str;		/* used in str_get and str_gnum macros */
	char *argstr = (char *)str_get(st[1]);

	st[0] = &str_undef;		/* default result on error is undef */

#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("do_sql: '%s'\n", argstr);
#endif

	/* If there isn't one around, allocate the maximum sqlda	*/
	/* Note: IISQLDA type has max size (IISQ_MAX_COLS cols) 	*/
	if (sqlda == (IISQLDA *)0)
	{
		if ( (sqlda=(IISQLDA *)malloc(sizeof(IISQLDA))) == NULL)
			fatal("ingperl: out of memory!");
		sqlda->sqln = IISQ_MAX_COLS; /* ingres can't give more */
	}

	/*	"connect dbname [-sqloptions]" (one space between each) */
	if (!strncmp(argstr, "connect ", 8))
	{
		EXEC SQL BEGIN DECLARE SECTION;
		char connectbuf[1024];
		char *connectopts = "";
		EXEC SQL END DECLARE SECTION;
		strcpy(connectbuf, &argstr[8]);	/* copy dbname and options */
		if ( (tmps = strchr(connectbuf,' ')) != NULL)
		{
			*tmps = '\0';				/* terminate string after dbname	*/
			connectopts = tmps+1;		/* point options at rest of string	*/
		}
		/* this is a just for Radhakrishnan radha@advtech.uswest.com	*/
		/* who had patched sqlperl to allow "indentified by"			*/
		if (!strncmp(connectopts, "identified by ", 14)){
			sprintf(connectopts, "-u%s", &connectopts[14]);
		}
		if (!connected && !getenv("II_SYSTEM"))
			warn("ingperl error: II_SYSTEM is not set. Ingres will fail.");

		EXEC SQL CONNECT :connectbuf OPTIONS = :connectopts;
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf("do_sql connect: %s (options=%s), sqlcode=%ld\n",
					connectbuf, connectopts, sqlca.sqlcode);
#endif
		if (sqlca.sqlcode != 0)
			return sp;
		st[0] = &str_yes;
		connected   = 1;
		cursor_open = 0;
    }

	else if (!strcmp(argstr, "disconnect"))
	{
		/* check that the app has commited or rolledback */
		EXEC SQL INQUIRE_INGRES(:SqlInt = TRANSACTION);
		if (SqlInt == 1){
			warn("ingperl: You should commit or rollback before disconnect.");
			warn("ingperl: Any outstanding changes have been rolledback.");
			EXEC SQL ROLLBACK;
			if (sqlca.sqlcode != 0)
				warn("ingperl: problem rolling back");
		}
		EXEC SQL DISCONNECT ALL;	/* ingres would commit by default */
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf("do_sql disconnect: sqlcode=%ld\n",sqlca.sqlcode);
#endif
		if (sqlca.sqlcode != 0)
			return sp;
	/*	Although the disconnect worked ok, Ingres seems to leave an	*/
	/*	internal error (E_LQ002E query issued outside of a session) */
	/*	in the error buffer! This means that $sql_error is not empty. */
		st[0] = &str_yes;
		connected   = 0;
		cursor_open = 0;
	}

	else if (!strcmp(argstr, "geterror"))	/* for compatibility */
	{
        st[0] = str_2mortal(str_make("",0));
		userval(UV_sql_error, st[0]); /* should use &sql_error now	*/
	}

	else if (!strcmp(argstr, "fetch"))		/* for compatibility */
		return do_sql_fetch(ix, sp, items);	/* should use &sql_fetch now	*/

	else	/* default to preparing anything we don't recognise */
	{
		if (!do_sql_prepare(argstr))
			st[0] = &str_yes;
    }
	return sp;
}


/* ===== Handlers (errors and warnings etc) ===== */

int
on_event(type)
int type;
{
    EXEC SQL BEGIN DECLARE SECTION;
        char    buf1[999];
        char    buf2[999];
    EXEC SQL END   DECLARE SECTION;
	char *event="error";
	buf1[0]=buf2[0]='\0';
#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("on_error %d: sqlcode=%ld\n",type, sqlca.sqlcode);
#endif
	switch(type){
	case OE_error:
		if (!sql_showerrors)
			return 0;
		EXEC SQL INQUIRE_INGRES(:buf1 = ERRORTEXT);
		break;
	case OE_warning:
		if (!sql_showerrors)
			return 0;
		event="warning";
		EXEC SQL INQUIRE_INGRES(:buf1 = ERRORTEXT);
		break;
	case OE_message:
		event="message";
		EXEC SQL INQUIRE_INGRES(:buf1 = MESSAGETEXT);
		break;
	case OE_dbevent:
		event="event";
		EXEC SQL INQUIRE_INGRES(:buf1=DBEVENTNAME, :buf2=DBEVENTTEXT);
		break;
	default:
		event="unknown!";
		break;
	}
	if (buf1[strlen(buf1)-1] == '\n')	/* chop trailing newline */
		buf1[strlen(buf1)-1] = '\0';
    printf("ingperl %s: %s %s\n", event, buf1, buf2);
	return 0;
}
int on_error()   { return on_event(OE_error);   }
int on_warning() { return on_event(OE_warning); }
int on_message() { return on_event(OE_message); }
int on_dbevent() { return on_event(OE_dbevent); }


/* ===== get and set variables ===== */

static int
userval(ix, str)
int ix;
STR *str;
{
	EXEC SQL BEGIN DECLARE SECTION;
	static char buf[500];
	EXEC SQL END DECLARE SECTION;
	int i;

    switch (ix) {

	case UV_sql_sqlcode:
		str_numset(str, (double)sqlca.sqlcode);
		break;

	case UV_sql_readonly:
		str_numset(str, (double)sql_readonly);
		break;

	case UV_sql_showerrors:
		str_numset(str, (double)sql_showerrors);
		break;

	case UV_sql_debug:
		str_numset(str, (double)sql_debug);
		break;

    case UV_sql_rowcount:
		EXEC SQL INQUIRE_INGRES(:SqlInt = ROWCOUNT);
		str_numset(str, (double)SqlInt);
		break;

    case UV_sql_error:
		EXEC SQL INQUIRE_INGRES(:buf = errortext);
		i = strlen(buf)-1;
		if (buf[i] == '\n')	/* chop trailing newline */
			buf[i] = '\0';
		str_set(str, buf);
		break;

	case UV_sql_version:
		sprintf(buf,"ingperl %d.%d pl%d (built %s %s)",
			IP_MAJOR, IP_MINOR, IP_PATCHLEVEL,
			/* if your compiler blows up on these then just	*/
			/* replace them with strings: "unknown","");	*/
			__DATE__,__TIME__);
		str_set(str, buf);
		break;

	default:
		warn("ingperl userval: unknown var");
		break;
	}
	return 0;
}
static int
userset(ix, str)
int ix;
STR *str;
{
	switch(ix){

	case UV_sql_readonly:
		sql_readonly = str_gnum(str);
		break;

	case UV_sql_showerrors:
		sql_showerrors = str_gnum(str);
		break;

	case UV_sql_debug:
		sql_debug = str_gnum(str);
#ifndef INGPERL_DEBUG
		if (sql_debug)
			printf("sql_debug: ingperl not compiled with INGPERL_DEBUG set\n");
#endif
		break;

    case UV_sql_error:
    case UV_sql_sqlcode:
    case UV_sql_version:
    case UV_sql_rowcount:
		warn("ingperl: cannot set read only variable");
		break;
	default:
		warn("ingperl userset: unknown var");
		break;
	}
	return 0;
}



/* returns undef on error else 1 or num rows affected if available */
int
do_sql_exec(ix, sp, items)
     int ix;
     register int sp;
     register int items;
{
	STR **st = stack->ary_array + sp;
	register STR *Str;		/* used in str_get and str_gnum macros */
	char *argstr = (char *)str_get(st[1]);

	SqlBuf = argstr;
	st[0] = &str_undef;
	EXEC SQL EXECUTE IMMEDIATE :SqlBuf;
#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("do_sql_exec: sqlcode=%ld\n",sqlca.sqlcode);
#endif
	if (sqlca.sqlcode >= 0)	/* statement ran ok */
		 st[0] = &str_yes;
	return sp;
}



int
do_sql_prepare(argstr)
	char *argstr;	/* sql string to prepare */
{
	static int SqlDataLen;
	static char *SqlData;

	int templen = 0;
	int i;

	SqlBuf = argstr;

	EXEC SQL PREPARE stmt FROM :SqlBuf;
	if (sqlca.sqlcode < 0)
		return 1;

	EXEC SQL DESCRIBE stmt INTO :sqlda;
	if (sqlca.sqlcode < 0)
		return 1;

	if (sqlda->sqld > sqlda->sqln)	/* should never happen */
		fatal("ingperl: too many rows!");

	/* sqlda->sqld will be zero if it wasn't a select statement.
	If that happens to be the case, just execute the statement
	and return an error code. */
	if (sqlda->sqld == 0)
	{
		EXEC SQL EXECUTE stmt;
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf("do_sql_prepare: executed, sqlcode=%ld\n",sqlca.sqlcode);
#endif
		if (sqlca.sqlcode < 0)
			return 1;
		else
			return 0;
	}

#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("do_sql_prepare: prepared, %d fields\n", sqlda->sqld);
#endif

	/* Otherwise, go down the sqlda allocating storage for things: */
	for (i = 0; i < sqlda->sqld; i++)
	{
		IISQLVAR *var = &sqlda->sqlvar[i];
		int origtype = var->sqltype;
		int type = origtype;
		int len  = 0;
		int ind  = 0;
		int baggage = 0;
		int align = 3;	/* 32 bit boundary, see float below */

		if (type < 0)	/* negative type indicates nullable	*/
		{
			ind = 1;
			type = -type;
		}

		switch (type)
	    {
	    case 30:	/* integer */
			len = sizeof(long);	/* allow for longest int */
			var->sqltype = ind ? -30 : 30;
			/* var->sqllen = sizeof(long); don't change len */
			break;

	    case 5:		/* money - treat as float8 */
			var->sqltype = ind ? -31 : 31;
			var->sqllen  = len = sizeof(double);
			align = 7;
			break;

	    case 31:	/* float 4 or 8 */
			len = var->sqllen;
			align = 7;
			break;

	    case 3:		/* date */
			len = IISQ_DTE_LEN;
			var->sqltype = ind ? -20 : 20;	/* treat date as char */
			var->sqllen  = IISQ_DTE_LEN;
			baggage = 1;
			break;

	    case 20:	/* char */
			len = var->sqllen;
			baggage = 1;
			break;

	    case 21:	/* varchar */
			len = var->sqllen + 2;
			baggage = 1;
			break;

		default:	/* oh dear! */
			len = var->sqllen + 8; baggage = 1;	/* take a guess */
			warn("ingperl: field %d has unsupported type %d",i+1,type);
			break;
		}

		var->sqlind = (short *)0;
		if (ind)
		{
			/* For sqlind, make sure we're on a 16-bit boundary: */
			if (templen & 1)
				templen += 1;
			var->sqlind = (short *)templen;
			templen += sizeof(short);	
		}

		/* For sqldata, make sure we're on a suitable boundary: */
		if (templen & align)
			templen = (templen + align) & ~align;

		var->sqldata = (char *)templen;
		templen += len + baggage;
#ifdef INGPERL_DEBUG
		if (sql_debug)
		{
			printf("do_sql_prepare F%d type=%3d (orig %3d), len=%2d (buf %d+%d)\n",
				i+1, var->sqltype, origtype, var->sqllen, len, baggage);
		}
#endif
	}

	/* Now that we've figured out how big everything is, let's make
	sure there's space for it: */
	if (templen > SqlDataLen || SqlData==NULL)
	{
		if (SqlData != (char *)0)
			free(SqlData);
		SqlData = (char *)malloc(templen);
		if (SqlData == (char *)0)
			return 1;
		SqlDataLen = templen;
	}

	/* now convert offsets into real pointers into SqlData memory */
	for (i=0; i < sqlda->sqld; i++)
	{
		IISQLVAR *var = &sqlda->sqlvar[i];
		if (var->sqltype < 0)
			var->sqlind = (short *)(SqlData + (int)(var->sqlind));

		var->sqldata = SqlData + (int)(var->sqldata);
	}

	if (cursor_open)
		close_sql_cursor();

	EXEC SQL DECLARE sql_cursor CURSOR FOR stmt;
	if (sqlca.sqlcode < 0)
		return 1;

	if (sql_readonly){
		EXEC SQL OPEN sql_cursor FOR READONLY;
	}else{
		EXEC SQL OPEN sql_cursor;
	}
	if (sqlca.sqlcode < 0)
		return 1;

	cursor_open = 1;
	return 0;

} /* End do_sql_prepare() */



static int
close_sql_cursor()
{
	/* Note that the cursor may be closed by ingres without us	*/
	/* being aware of it. In which case we may get spurious		*/
	/* errors when closing. */
	if (!cursor_open)
	{
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf("do_sql_close: ignored, cursor not open.\n");
#endif
		return(0);
	}
	EXEC SQL CLOSE sql_cursor;
	cursor_open = 0;
#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("do_sql_close: sqlcode=%ld\n", sqlca.sqlcode);
#endif
	return sqlca.sqlcode;
}


/* returns undef on error else 1 or num rows affected if available */
int
do_sql_close(ix, sp, items)
     int ix;
     register int sp;
     register int items;
{
	STR **st = stack->ary_array + sp;

	close_sql_cursor();	/* just return undef for now */

	st[0] = &str_yes;
	return sp;
}




int
do_sql_desc(ix, sp, items)
	 int ix;
     register int sp;
     register int items;
{
	STR **st = stack->ary_array + sp;
	register int i;
	char *type;
	astore(stack, sp + sqlda->sqld, Nullstr);
	st = stack->ary_array + sp;
	for (i=0; i < sqlda->sqld; i++)
	{
		IISQLVAR *var = &sqlda->sqlvar[i];
		switch(ix){
		case US_sql_ingtypes:
			st[i] = str_2mortal(str_make("",0));
			str_numset(st[i], (double)abs(var->sqltype));
			break;
		case US_sql_nullable:	/* 0=not nullable, 1=nullable */
			st[i] = str_2mortal(str_make("",0));
			str_numset(st[i], (double)(var->sqltype < 0));
			break;
		case US_sql_types:		/* 'd', 'f', 's'	*/
			switch(abs(var->sqltype)){
			case 30:	type="d";	break;
			case 31: case 5: case 10:
						type="f";	break;
			default:	type="s";	break;
			}
			st[i] = str_2mortal(str_make(type,strlen(type)));
			break;
		case US_sql_lengths:
			st[i] = str_2mortal(str_make("",0));
			str_numset(st[i], (double)(var->sqllen));
			break;
		case US_sql_names:
			st[i] = str_2mortal(
				str_make(var->sqlname.sqlnamec,var->sqlname.sqlnamel));
			break;
		default:
			warn("do_sql_desc: bad func!");
			break;
		}
    }
	return sp + sqlda->sqld - 1;
}




int
do_sql_fetch(ix, sp, items)
	 int ix;
     register int sp;
     register int items;
{
	STR **st = stack->ary_array + sp;
	register int i;
	register char *tmps;

#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf("do_sql_fetch: ");
#endif
	EXEC SQL FETCH sql_cursor USING DESCRIPTOR :sqlda;

	if (sqlca.sqlcode != 0)
    {
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf("sqlcode=%ld\n", sqlca.sqlcode);
#endif
		close_sql_cursor();
		/* return empty array */
		astore(stack, sp + 0, Nullstr);
		st = stack->ary_array + sp;
		return sp + 0 - 1;
	}
  
	astore(stack, sp + sqlda->sqld, Nullstr);
	st = stack->ary_array + sp;

	for (i = 0; i < sqlda->sqld; i++)
	{
		int type;
		IISQLVAR *var = &sqlda->sqlvar[i];
		union _data {
			char *c;
			double *df;
			float  *sf;
			struct {
				short l;
				char c;
			} *v;
			long  *i4;
			short *i2;
			char  *i1;
		} data;
      
		data.c = var->sqldata;
				  
#ifdef INGPERL_DEBUG
		if (sql_debug)
			printf( (i==0) ? "(%d:" : ", %d:", i);
#endif

		if (var->sqltype < 0 && *(var->sqlind) < 0)
		{
			st[i] = str_2mortal(&str_undef);
#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("<NULL>");
#endif
			continue;
		}

		if (var->sqltype < 0)
			 type = -var->sqltype;
		else type =  var->sqltype;

		switch (type)
		{
	    case 30:	/* int */
			st[i] = str_mortal(&str_no);
			switch(var->sqllen)
			{
			case 4:	str_numset(st[i], (double)(*data.i4)); break;
			case 2:	str_numset(st[i], (double)(*data.i2)); break;
			case 1:	str_numset(st[i], (double)(*data.i1)); break;
			}
#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("%g ", (double)str_gnum(st[i]));
#endif
			break;
		      
	    case 31:	/* float */
			st[i] = str_mortal(&str_no);
			if (var->sqllen == sizeof(float))
				 str_numset(st[i], (double)(*data.sf));
			else str_numset(st[i], *data.df);

#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("%g ", (double)str_gnum(st[i]));
#endif
			break;
		      
	    case 20:	/* char */
			data.c[var->sqllen] = 0;	/* do we need to do this ? */
			tmps = data.c;
			st[i] = str_2mortal(str_make(tmps, var->sqllen));
#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("'%s' ", data.c);
#endif
			break;

	    case 21:	/* varchar */
			(&(data.v->c))[data.v->l] = 0;	/* do we need to do this ? */
			tmps = &(data.v->c);
			st[i] = str_2mortal(str_make(tmps, data.v->l));
#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("'%s' ", &(data.v->c));
#endif
			break;

		default:	/* unhandled type, warn()'d during prepare */
			st[i] = str_2mortal(&str_undef);
#ifdef INGPERL_DEBUG
			if (sql_debug)
				printf("<badtype%d> ",type);
#endif
			break;
		}
    }

#ifdef INGPERL_DEBUG
	if (sql_debug)
		printf(")\n");
#endif

	return sp + sqlda->sqld - 1;
}


/* end */
